from dataclasses import dataclass
from typing import (
    Optional,
    List,
    Iterator,
    Iterable,
    Union,
    Sized,
    Dict,
    Any,
    Type,
    Callable,
)

from marshy.types import ExternalType, ExternalItemType

from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass(frozen=True)
class AnyOfSchema(SchemaABC):
    schemas: Union[Iterable[SchemaABC], Sized]
    name: str = None
    description: str = None

    def __post_init__(self):
        schemas = []
        for s in self.schemas:
            if isinstance(s, AnyOfSchema):
                schemas.extend(s.schemas)
            else:
                schemas.append(s)
        object.__setattr__(self, "schemas", tuple(schemas))

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        errors = [SchemaError(current_path or [], "type", item)]
        for schema in self.schemas:
            errors = list(schema.get_schema_errors(item, current_path))
            if not errors:
                return
        if item is not None:
            yield from errors

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        any_of = [s.dump_json_schema(json_context) for s in self.schemas]
        dumped = dict(anyOf=any_of)
        if self.name:
            dumped["name"] = self.name
        if self.description:
            dumped["description"] = self.description
        return dumped

    def simplify(self) -> SchemaABC:
        any_of = [s.simplify() for s in self.schemas]
        schema = AnyOfSchema(**{**self.__dict__, "schemas": any_of})
        return schema

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        type_ = existing_types.get(self.name)
        if type_:
            return type_
        type_ = Union[
            tuple(
                s.get_normalized_type(existing_types, object_wrapper)
                for s in self.schemas
            )
        ]
        existing_types[self.name] = type_
        return type_



from dataclasses import dataclass
from typing import Optional, List, Iterator, Dict, Tuple, Union, Any, Type, Callable

from marshy.types import ExternalItemType, ExternalType

from schemey.json_schema_context import JsonSchemaContext
from schemey.optional_schema import NoDefault
from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC
from schemey.schema_error import SchemaError

_instance = None


@dataclass(frozen=True)
class BooleanSchema(SchemaABC):
    description: str = None

    def get_schema_errors(
        self, item: bool, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if not isinstance(item, bool):
            yield SchemaError(current_path or [], "type", item)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        dumped = dict(type="boolean")
        if self.description:
            dumped["description"] = self.description
        return dumped

    def get_param_schemas(self, current_path: str) -> List[ParamSchema]:
        return [ParamSchema(name=current_path, schema=self)]

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> Union[ExternalType, NoDefault]:
        if current_path not in params:
            return NoDefault
        values = params.get(current_path)
        value = values[0].lower() not in ["", "0", "0.0", "false"]
        return value

    def to_url_params(self, current_path: str, item: bool) -> Iterator[Tuple[str, str]]:
        value = "1" if item else "0"
        yield current_path, value

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type[bool]:
        return bool



from typing import Optional

from marshy.types import ExternalItemType

from schemey.enum_schema import EnumSchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class EnumSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if "enum" not in item or "name" not in item:
            return None
        loaded = EnumSchema(
            name=item.get("name"),
            enum=set(item.get("enum")),
            description=item.get("description"),
        )
        return loaded



import json

import marshy

from tests.examples.a_hello_world import (
    schema,
)  # Continuing from the previous example...

schema_json = marshy.dump(schema.json_schema)  # Convert to json_schema using marshy
json_str = json.dumps(schema_json)  # Dump to a string using the standard library

assert schema_json == {
    "type": "object",
    "name": "HelloWorld",
    "additionalProperties": False,
    "properties": {
        "name": {"type": "string"},
        "age": {"type": "integer", "default": None},
        "friend": {"type": "boolean", "default": None},
    },
    "required": ["name"],
}



from typing import Optional

from marshy.types import ExternalItemType

from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.loader.schema_loader_abc import SchemaLoaderABC


class RefSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if "$ref" not in item:
            return None
        ref = item["$ref"]
        if not ref.startswith(json_context.defs_path):
            return None  # We don't support loading from anywhere in the document for now...
        ref = ref[len(json_context.defs_path) :]
        schema = json_context.defs[ref]
        schema.num_usages += 1
        return schema



from datetime import datetime
from uuid import UUID

from schemey.boolean_schema import BooleanSchema
from schemey.factory.any_of_schema_factory import AnyOfSchemaFactory
from schemey.factory.array_schema_factory import ArraySchemaFactory
from schemey.factory.dataclass_schema_factory import DataclassSchemaFactory
from schemey.factory.enum_schema_factory import EnumSchemaFactory
from schemey.factory.factory_schema_factory import FactorySchemaFactory
from schemey.factory.impl_schema_factory import ImplSchemaFactory
from schemey.factory.tuple_schema_factory import TupleSchemaFactory
from schemey.integer_schema import IntegerSchema
from schemey.loader.any_of_schema_loader import AnyOfSchemaLoader
from schemey.loader.array_schema_loader import ArraySchemaLoader
from schemey.loader.boolean_schema_loader import BooleanSchemaLoader
from schemey.loader.const_schema_loader import ConstSchemaLoader
from schemey.loader.enum_schema_loader import EnumSchemaLoader
from schemey.loader.integer_schema_loader import IntegerSchemaLoader
from schemey.loader.null_schema_loader import NullSchemaLoader
from schemey.loader.number_schema_loader import NumberSchemaLoader
from schemey.loader.object_schema_loader import ObjectSchemaLoader
from schemey.loader.ref_schema_loader import RefSchemaLoader
from schemey.loader.string_schema_loader import StringSchemaLoader
from schemey.loader.tuple_schema_loader import TupleSchemaLoader
from schemey.null_schema import NullSchema
from schemey.number_schema import NumberSchema
from schemey.schema_context import SchemaContext
from schemey.string_format import StringFormat
from schemey.string_schema import StringSchema

priority = 100


def configure(context: SchemaContext):
    configure_schemas(context)
    configure_factories(context)
    configure_loaders(context)


def configure_schemas(context: SchemaContext):
    context.register_schema(bool, BooleanSchema())
    context.register_schema(int, IntegerSchema())
    context.register_schema(type(None), NullSchema())
    context.register_schema(float, NumberSchema())
    context.register_schema(str, StringSchema())
    context.register_schema(datetime, StringSchema(format=StringFormat.DATE_TIME))
    context.register_schema(UUID, StringSchema(format=StringFormat.UUID))


def configure_factories(context: SchemaContext):
    context.register_factory(EnumSchemaFactory())
    context.register_factory(ArraySchemaFactory())
    context.register_factory(TupleSchemaFactory())
    context.register_factory(DataclassSchemaFactory())
    context.register_factory(EnumSchemaFactory())
    context.register_factory(FactorySchemaFactory())
    context.register_factory(ImplSchemaFactory())
    context.register_factory(AnyOfSchemaFactory())


def configure_loaders(context: SchemaContext):
    context.register_loader(AnyOfSchemaLoader())
    context.register_loader(ArraySchemaLoader())
    context.register_loader(BooleanSchemaLoader())
    context.register_loader(ConstSchemaLoader())
    context.register_loader(EnumSchemaLoader())
    context.register_loader(IntegerSchemaLoader())
    context.register_loader(NullSchemaLoader())
    context.register_loader(NumberSchemaLoader())
    context.register_loader(ObjectSchemaLoader())
    context.register_loader(RefSchemaLoader())
    context.register_loader(StringSchemaLoader())
    context.register_loader(TupleSchemaLoader())



from abc import ABC
from typing import List, Union, Dict, Iterator, Tuple

from marshy import ExternalType

from schemey.optional_schema import NoDefault
from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC


class StrParamSchemaABC(SchemaABC, ABC):
    def get_param_schemas(self, current_path: str) -> List[ParamSchema]:
        return [ParamSchema(name=current_path, schema=self)]

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> Union[ExternalType, NoDefault]:
        if current_path not in params:
            return NoDefault
        values = params.get(current_path)
        return values[0]

    def to_url_params(self, current_path: str, item: str) -> Iterator[Tuple[str, str]]:
        yield current_path, str(item)



import copy
from typing import Optional, Set

from marshy.types import ExternalItemType

from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.object_schema import ObjectSchema
from schemey.optional_schema import OptionalSchema, NoDefault


class ObjectSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if item.get("type") != "object":
            return None
        name = item.get("name")
        properties = item.get("properties") or {}
        required = set(item.get("required") or [])
        properties = {
            k: self._load_optional(k, s, required, json_context)
            for k, s in properties.items()
        }
        loaded = ObjectSchema(
            name=name,
            properties=properties,
            additional_properties=item.get("additionalProperties"),
            required=required or None,
            description=item.get("description"),
        )
        return loaded

    @staticmethod
    def _load_optional(
        key: str, item: ExternalItemType, required: Set[str], context: JsonSchemaContext
    ) -> SchemaABC:
        schema = context.load(item)
        if key not in required:
            default = NoDefault
            if "default" in item:
                default = copy.deepcopy(item["default"])
            schema = OptionalSchema(schema, default)
        return schema



from dataclasses import dataclass
from typing import Optional

from schemey.schema_context import schema_for_type


@dataclass
class HelloWorld:
    name: str
    age: Optional[int] = None
    friend: Optional[bool] = None


# A schema may be generated...
schema = schema_for_type(HelloWorld)

# A full listing of errors for an item can be retrieved:
# noinspection PyTypeChecker
errors = list(
    schema.get_schema_errors(HelloWorld(None))
)  # [SchemaError(path='name', code='type', value='You')]

# ...or validation errors can be thrown:
schema.validate(HelloWorld("You"))  # No errors
# schema.validate(HelloWorld(None)) # raises SchemaError(path='name', code='type', value=None)



from typing import Optional

from marshy.types import ExternalItemType

from schemey.boolean_schema import BooleanSchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class BooleanSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if item.get("type") != "boolean":
            return None
        loaded = BooleanSchema(item.get("description"))
        return loaded



from typing import Type, Optional, Set, List, Tuple, Union

import typing_inspect
from marshy.utils import resolve_forward_refs

from schemey.array_schema import ArraySchema
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class ArraySchemaFactory(SchemaFactoryABC):
    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        array_type = self.get_array_type(type_)
        if array_type:
            args = typing_inspect.get_args(type_)
            item_type = resolve_forward_refs(args[0])
            schema = json_context.get_schema(item_type)
            uniqueness = array_type is Set
            return ArraySchema(item_schema=schema, uniqueness=uniqueness)

    @staticmethod
    def get_array_type(type_: Type) -> Union[Type[List], Type[Set], Type[Tuple], None]:
        origin = typing_inspect.get_origin(type_)
        if origin is list:
            return List
        if origin is set:
            return Set
        if origin is tuple:
            args = typing_inspect.get_args(type_)
            if len(args) == 2 and args[1] is Ellipsis:
                return Tuple



from typing import Optional

from marshy.types import ExternalItemType

from schemey._util import filter_none
from schemey.array_schema import ArraySchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class ArraySchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if item.get("type") != "array":
            return None
        item_schema = None
        if "items" in item:
            item_schema = json_context.load(item.get("items"))
        loaded = ArraySchema(
            **filter_none(
                dict(
                    item_schema=item_schema,
                    min_items=item.get("minItems"),
                    max_items=item.get("maxItems"),
                    uniqueness=item.get("uniqueness"),
                    description=item.get("description"),
                )
            )
        )
        return loaded



from dataclasses import dataclass, MISSING
from typing import Optional, Type, List, Iterator, TypeVar, Dict, Union, Tuple

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

from schemey.optional_schema import NoDefault
from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC
from schemey.schema_error import SchemaError

T = TypeVar("T")


@dataclass
class ObjSchema:
    """Implementation of SchemaABC using Json"""

    json_schema: SchemaABC
    marshaller: MarshallerABC[T]

    def get_schema_errors(
        self, item: T, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        try:
            dumped = self.marshaller.dump(item)
        except (ValueError, AttributeError):
            yield SchemaError(current_path, "type", item)
            return
        yield from self.json_schema.get_schema_errors(dumped, current_path)

    @property
    def item_type(self) -> Type[T]:
        return self.marshaller.marshalled_type

    def validate(self, item: T, current_path: Optional[List[str]] = None):
        """Validate the item given"""
        errors = self.get_schema_errors(item, current_path)
        error = next(errors, None)
        if error:
            raise error

    def get_param_schemas(self, current_path: str = "") -> Optional[List[ParamSchema]]:
        """Optional schemas allow only one parameter"""
        return self.json_schema.get_param_schemas(current_path)

    def from_url_params(
        self, params: Dict[str, List[str]], current_path: str = ""
    ) -> Union[ExternalType, NoDefault]:
        item = self.json_schema.from_url_params(current_path, params)
        if item is MISSING:
            return None
        elif item is NoDefault:
            raise ValueError("missing_required_fields")
        loaded = self.marshaller.load(item)
        return loaded

    def to_url_params(
        self, item: T, current_path: str = ""
    ) -> Iterator[Tuple[str, str]]:
        dumped = self.marshaller.dump(item)
        yield from self.json_schema.to_url_params(current_path, dumped)



from typing import Optional

from marshy.types import ExternalItemType

from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.string_schema import StringSchema
from schemey.string_format import StringFormat


class StringSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if item.get("type") != "string":
            return None
        loaded = StringSchema(
            min_length=item.get("minLength"),
            max_length=item.get("maxLength"),
            pattern=item.get("pattern"),
            format=StringFormat(item.get("format")) if item.get("format") else None,
            description=item.get("description"),
        )
        return loaded



from dataclasses import dataclass
from typing import Optional, List, Iterator, Union, Dict, Any, Type, Callable

from marshy.types import ExternalItemType

from schemey._util import filter_none
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError
from schemey.str_param_schema_abc import StrParamSchemaABC


@dataclass(frozen=True)
class IntegerSchema(StrParamSchemaABC):
    minimum: Optional[int] = None
    exclusive_minimum: Optional[int] = None
    maximum: Optional[int] = None
    exclusive_maximum: Optional[int] = None
    description: Optional[str] = None

    def get_schema_errors(
        self, item: int, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if not isinstance(item, int):
            yield SchemaError(current_path, "type", item)
            return
        yield from check(
            item,
            current_path,
            self.minimum,
            self.maximum,
            self.exclusive_minimum,
            self.exclusive_maximum,
        )

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        return dump_json_schema(self, "integer")

    def simplify(self) -> SchemaABC:
        kwargs = simplify_kwargs(self)
        return IntegerSchema(**kwargs)

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type[int]:
        return int


def simplify_kwargs(schema):
    kwargs = {**schema.__dict__}
    if schema.minimum is not None and schema.exclusive_minimum is not None:
        del kwargs[
            "minimum"
            if schema.minimum <= schema.exclusive_minimum
            else "exclusive_minimum"
        ]
    if schema.maximum is not None and schema.exclusive_maximum is not None:
        del kwargs[
            "maximum"
            if schema.maximum >= schema.exclusive_maximum
            else "exclusive_maximum"
        ]
    return kwargs


def dump_json_schema(item, type_name: str) -> Optional[ExternalItemType]:
    dumped = filter_none(
        dict(
            type=type_name,
            minimum=item.minimum,
            exclusiveMinimum=item.exclusive_minimum,
            maximum=item.maximum,
            exclusiveMaximum=item.exclusive_maximum,
            description=item.description,
        )
    )
    return dumped


def check(
    item: Union[float, int],
    current_path,
    minimum,
    maximum,
    exclusive_minimum,
    exclusive_maximum,
):
    if minimum is not None and item < minimum:
        yield SchemaError(current_path, "minimum", item)
    if exclusive_minimum is not None and item <= exclusive_minimum:
        yield SchemaError(current_path, "exclusive_minimum", item)
    if maximum is not None and item > maximum:
        yield SchemaError(current_path, "maximum", item)
    if exclusive_maximum is not None and item >= exclusive_maximum:
        yield SchemaError(current_path, "exclusive_maximum", item)



from typing import Optional

from marshy.types import ExternalItemType

from schemey.integer_schema import IntegerSchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class IntegerSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if item.get("type") != "integer":
            return None
        loaded = IntegerSchema(
            minimum=item.get("minimum"),
            exclusive_minimum=item.get("exclusiveMinimum"),
            maximum=item.get("maximum"),
            exclusive_maximum=item.get("exclusiveMaximum"),
            description=item.get("description"),
        )
        return loaded



import copy
from dataclasses import dataclass, MISSING
from typing import Optional, List, Iterator, Union, Type, Tuple, Dict, Any, Callable

from marshy import ExternalType
from marshy.factory.optional_marshaller_factory import get_optional_type
from marshy.types import ExternalItemType

from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC, NoDefault
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass
class OptionalSchema(SchemaABC):
    """
    Attributes can be optional or required. ObjectSchemas contain a list of required property names,
    and all other properties should be Optional - and may have a default value (NoDefault is used
    for dynamic defaults - e.g.: field(default_factory=uuid4)
    """

    schema: SchemaABC
    default: Union[ExternalType, Type[NoDefault]] = NoDefault

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if item is not None and item is not MISSING:
            yield from self.schema.get_schema_errors(item, current_path)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        schema = self.schema.dump_json_schema(json_context)
        if self.default is not NoDefault:
            schema["default"] = copy.deepcopy(self.default)
        return schema

    def simplify(self) -> SchemaABC:
        schema = OptionalSchema(self.schema.simplify(), self.default)
        return schema

    def get_param_schemas(self, current_path: str) -> Optional[List[ParamSchema]]:
        """Optional schemas allow only one parameter"""
        schemas = self.schema.get_param_schemas(current_path)
        if schemas and len(schemas) == 1:
            schemas[0].required = False
            return schemas

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> Union[ExternalType, type(MISSING)]:
        params = self.schema.from_url_params(current_path, params)
        if params is NoDefault:
            if self.default is NoDefault:
                params = MISSING
            else:
                params = copy.deepcopy(self.default)
        return params

    def to_url_params(
        self, current_path: str, item: ExternalType
    ) -> Iterator[Tuple[str, str]]:
        yield from self.schema.to_url_params(current_path, item)

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        type_ = self.schema.get_normalized_type(existing_types, object_wrapper)
        if self.default is None:
            type_ = Optional[get_optional_type(type_) or type_]
        return type_



from dataclasses import dataclass, field
from typing import Type

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from schemey.deferred_schema import DeferredSchema
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_context import SchemaContext, get_default_schema_context


@dataclass(frozen=True)
class SchemaMarshaller(MarshallerABC[SchemaABC]):
    marshalled_type: Type[SchemaABC] = SchemaABC
    schemey_context: SchemaContext = field(default_factory=get_default_schema_context)
    defs_key: str = "$defs"

    def load(self, item: ExternalItemType) -> SchemaABC:
        raw_defs = item.get(self.defs_key) or {}
        defs = {k: DeferredSchema(k) for k in raw_defs}
        json_schema_context = JsonSchemaContext(
            defs=defs,
            loaders=self.schemey_context.schema_loaders,
            defs_path=f"#{self.defs_key}/",
        )
        for k, raw_def in raw_defs.items():
            deferred = defs[k]
            deferred.schema = json_schema_context.load(raw_def)
        schema = json_schema_context.load(item)
        schema = schema.simplify()
        return schema

    def dump(self, item: SchemaABC) -> ExternalType:
        json_context = JsonSchemaContext()
        dumped = item.dump_json_schema(json_context)
        if json_context.defs:
            dumped[self.defs_key] = {
                k: s.schema.dump_json_schema(json_context)
                for k, s in json_context.defs.items()
                if s.num_usages > 1
            }
        return dumped



from abc import abstractmethod, ABC
from functools import total_ordering
from typing import Optional

from marshy.types import ExternalItemType

from schemey.schema_abc import SchemaABC

_JsonSchemaContext = "schemey.json_schema_context.JsonSchemaContext"


@total_ordering
class SchemaLoaderABC(ABC):
    @abstractmethod
    def load(
        self, item: ExternalItemType, json_context: _JsonSchemaContext
    ) -> Optional[SchemaABC]:
        """Load a schema - return None if not possible using this jsonifier"""

    @property
    def priority(self) -> int:
        return 0

    def __ne__(self, other):
        return self.priority != getattr(other, "priority", None)

    def __lt__(self, other):
        return self.priority < getattr(other, "priority", None)



from typing import Optional

from marshy.types import ExternalItemType

from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.null_schema import NullSchema


class NullSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        loaded = NullSchema() if item.get("type") == "null" else None
        return loaded



from enum import Enum
from inspect import isclass
from typing import Type, Optional

from schemey.enum_schema import EnumSchema
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class EnumSchemaFactory(SchemaFactoryABC):
    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if isclass(type_) and issubclass(type_, Enum):
            # noinspection PyTypeChecker
            return EnumSchema(enum={e.value for e in type_}, name=type_.__name__)



from dataclasses import dataclass
from typing import Any, Union, List

from marshy import ExternalType


@dataclass(frozen=True)
class SchemaError(Exception):
    path: str
    code: str
    value: Any

    def __init__(
        self, path: Union[str, List[str]], code: str, value: ExternalType = None
    ):
        if not isinstance(path, str):
            path = "/".join(path or [])
        super().__init__(f":{code}:{path}:{value}")
        object.__setattr__(self, "path", path)
        object.__setattr__(self, "code", code)
        object.__setattr__(self, "value", value)



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Type

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC

_DeferredSchema = "schemey.deferred_schema.DeferredSchema"


@dataclass
class JsonSchemaContext:
    defs: Dict[str, _DeferredSchema] = field(default_factory=dict)
    defs_path: str = "#$defs"
    schemas: Optional[Dict[Type, SchemaABC]] = field(default_factory=dict)
    loaders: Optional[List[SchemaLoaderABC]] = None
    factories: Optional[List[SchemaFactoryABC]] = None
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)

    def load(self, item: ExternalItemType) -> SchemaABC:
        for loader in self.loaders:
            schema = loader.load(item, self)
            if schema:
                return schema
        raise ValueError(f"could_not_load:{item}")

    def get_schema(self, item_type: Type) -> SchemaABC:
        schema = self.schemas.get(item_type)
        if schema:
            return schema
        for factory in self.factories:
            schema = factory.create(item_type, self)
            if schema:
                self.schemas[item_type] = schema
                return schema
        raise ValueError(f"could_not_create:{item_type}")



from marshy.marshaller_context import MarshallerContext

from schemey.factory.schema_marshaller_factory import SchemaMarshallerFactory

priority = 100


def configure(context: MarshallerContext):
    context.register_factory(SchemaMarshallerFactory())



from typing import Type, Optional, Union

import typing_inspect

from schemey.any_of_schema import AnyOfSchema
from schemey.const_schema import ConstSchema
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.tuple_schema import TupleSchema


class AnyOfSchemaFactory(SchemaFactoryABC):
    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        origin = typing_inspect.get_origin(type_)
        if origin == Union:
            args = typing_inspect.get_args(type_)
            schemas = tuple(
                TupleSchema((ConstSchema(a.__name__), json_context.get_schema(a)))
                for a in args
            )
            return AnyOfSchema(schemas)



from dataclasses import dataclass
from typing import Type, Optional, Set

from marshy.factory.impl_marshaller_factory import ImplMarshallerFactory

from schemey.any_of_schema import AnyOfSchema
from schemey.const_schema import ConstSchema
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.tuple_schema import TupleSchema


@dataclass
class ImplSchemaFactory(SchemaFactoryABC):
    priority: int = 150

    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        impls = self.get_impls(type_, json_context)
        if impls:
            schemas = tuple(self.wrap_impl(i, json_context) for i in impls)
            return AnyOfSchema(schemas=schemas, name=type_.__name__)

    @staticmethod
    def get_impls(type_: Type, json_context: JsonSchemaContext) -> Optional[Set[Type]]:
        factories = json_context.marshaller_context.get_factories()
        for factory in factories:
            if isinstance(factory, ImplMarshallerFactory):
                if factory.base == type_:
                    return factory.impls

    @staticmethod
    def wrap_impl(type_: Type, json_context: JsonSchemaContext):
        schema = json_context.get_schema(type_)
        schema = TupleSchema((ConstSchema(type_.__name__), schema))
        return schema



from dataclasses import dataclass
from typing import Optional, List, Iterator, Dict, Any, Type, Callable

from marshy.types import ExternalItemType

from schemey.integer_schema import check, dump_json_schema, simplify_kwargs
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_abc import SchemaABC
from schemey.schema_error import SchemaError
from schemey.str_param_schema_abc import StrParamSchemaABC


@dataclass(frozen=True)
class NumberSchema(StrParamSchemaABC):
    minimum: Optional[float] = None
    exclusive_minimum: Optional[float] = None
    maximum: Optional[float] = None
    exclusive_maximum: Optional[float] = None
    description: Optional[str] = None

    def get_schema_errors(
        self, item: float, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if not isinstance(item, float) and item.__class__ is not int:
            yield SchemaError(current_path or [], "type", item)
            return
        yield from check(
            item,
            current_path,
            self.minimum,
            self.maximum,
            self.exclusive_minimum,
            self.exclusive_maximum,
        )

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        return dump_json_schema(self, "number")

    def simplify(self) -> SchemaABC:
        kwargs = simplify_kwargs(self)
        return NumberSchema(**kwargs)

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type[float]:
        return float



from dataclasses import dataclass
from datetime import datetime, time
import re
from typing import Optional, List, Iterator, Dict, Any, Union, Type, Callable
from uuid import UUID

import validators
from marshy.types import ExternalItemType

from schemey._util import filter_none
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError
from schemey.str_param_schema_abc import StrParamSchemaABC
from schemey.string_format import StringFormat


@dataclass(frozen=True)
class StringSchema(StrParamSchemaABC):
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    pattern: Optional[str] = None
    format: Optional[StringFormat] = None
    _compiled_pattern = None
    description: Optional[str] = None

    def __post_init__(self):
        object.__setattr__(
            self,
            "_compiled_pattern",
            None if self.pattern is None else re.compile(self.pattern),
        )

    def get_schema_errors(
        self, item: str, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if not isinstance(item, str):
            yield SchemaError(current_path, "type", item)
            return
        if self.min_length is not None and len(item) < self.min_length:
            yield SchemaError(current_path, "min_length", item)
        if self.max_length is not None and len(item) > self.max_length:
            yield SchemaError(current_path, "max_length", item)
        if self._compiled_pattern is not None and not self._compiled_pattern.search(
            item
        ):
            yield SchemaError(current_path, "pattern", item)
        if self.format == StringFormat.DATE:
            try:
                if datetime.fromisoformat(item).isoformat()[:10] != item:
                    yield SchemaError(current_path, "format:date", item)
            except ValueError:
                yield SchemaError(current_path, "format:date", item)
        elif self.format == StringFormat.DATE_TIME:
            try:
                datetime.fromisoformat(item).isoformat()
            except ValueError:
                yield SchemaError(current_path, "format:date-time", item)
        elif self.format == StringFormat.EMAIL:
            if validators.email(item) is not True:
                yield SchemaError(current_path, "format:email", item)
        elif self.format == StringFormat.HOSTNAME:
            if validators.domain(item) is not True:
                yield SchemaError(current_path, "format:hostname", item)
        elif self.format == StringFormat.IPV4:
            if validators.ipv4(item) is not True:
                yield SchemaError(current_path, "format:ipv4", item)
        elif self.format == StringFormat.IPV6:
            if validators.ipv6(item) is not True:
                yield SchemaError(current_path, "format:ipv6", item)
        elif self.format == StringFormat.TIME:
            try:
                time.fromisoformat(item).isoformat()
            except ValueError:
                yield SchemaError(current_path, "format:time", item)
        elif self.format == StringFormat.URI:
            if validators.url(item) is not True:
                yield SchemaError(current_path, "format:uri", item)
        elif self.format == StringFormat.UUID:
            if validators.uuid(item) is not True:
                yield SchemaError(current_path, "format:uuid", item)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        dumped = filter_none(
            dict(
                type="string",
                minLength=self.min_length,
                maxLength=self.max_length,
                pattern=self.pattern,
                format=self.format.value if self.format else None,
                description=self.description,
            )
        )
        return dumped

    def get_normalized_type(
        self, existing_types: Dict[str, Any], obj_wrapper: Callable
    ) -> Union[Type[str], Type[UUID], Type[datetime]]:
        if self.format == StringFormat.DATE_TIME:
            return datetime
        elif self.format == StringFormat.UUID:
            return UUID
        return str



import dataclasses
from typing import Type, Optional, Tuple

from marshy.factory.optional_marshaller_factory import get_optional_type

from schemey.deferred_schema import DeferredSchema
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.object_schema import ObjectSchema
from schemey.optional_schema import OptionalSchema, NoDefault


class DataclassSchemaFactory(SchemaFactoryABC):
    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if not dataclasses.is_dataclass(type_):
            return
        name = type_.__name__
        schema = json_context.defs.get(name)
        if schema:
            schema.num_usages += 1
            return schema
        schema = DeferredSchema(ref=name, num_usages=1)
        json_context.defs[name] = schema
        # noinspection PyDataclass
        fields = dataclasses.fields(type_)
        field_schemas = tuple(self._schema_for_field(f, json_context) for f in fields)
        required = {n for n, s in field_schemas if not isinstance(s, OptionalSchema)}
        schema.schema = ObjectSchema(
            properties={n: s for n, s in field_schemas},
            name=type_.__name__,
            required=required or None,
        )
        return schema

    @staticmethod
    def _schema_for_field(
        field: dataclasses.Field, json_context: JsonSchemaContext
    ) -> Tuple[str, SchemaABC]:
        schema = field.metadata.get("schemey")
        field_type = get_optional_type(field.type) or field.type
        if not schema:
            schema = json_context.get_schema(field_type)
        default = NoDefault
        if field.default is not dataclasses.MISSING:
            if field.default is None:
                default = None
            else:
                default = json_context.marshaller_context.dump(
                    field.default, field_type
                )
        if default is not NoDefault or field.default_factory is not dataclasses.MISSING:
            schema = OptionalSchema(schema, default)
        return field.name, schema



from dataclasses import dataclass
from typing import Optional, List, Iterator, Tuple, Dict, Type, Callable, Any

from marshy import ExternalType
from marshy.types import ExternalItemType

from schemey.object_schema import build_attributes
from schemey.optional_schema import NoDefault
from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC, _JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass
class TupleSchema(SchemaABC):
    schemas: Tuple[SchemaABC, ...]
    description: Optional[str] = None

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if current_path is None:
            current_path = []
        if not isinstance(item, list) or len(item) != len(self.schemas):
            yield SchemaError(current_path, "type", item)
            return
        for index, sub_item in enumerate(item):
            current_path.append(str(index))
            schema = self.schemas[index]
            yield from schema.get_schema_errors(sub_item, current_path)
            current_path.pop()

    def dump_json_schema(self, json_context: _JsonSchemaContext) -> ExternalItemType:
        dumped = {
            "type": "array",
            "prefixItems": [i.dump_json_schema(json_context) for i in self.schemas],
            "items": False,
        }
        if self.description:
            dumped["description"] = self.description
        return dumped

    def get_param_schemas(self, current_path: str) -> Optional[List[ParamSchema]]:
        param_schemas = []
        for index, schema in enumerate(self.schemas):
            sub_path = self._sub_path(current_path, index)
            sub_schemas = schema.get_param_schemas(sub_path)
            if sub_schemas is None:
                return None
            param_schemas.extend(sub_schemas)
        return param_schemas

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> ExternalType:
        values = []
        for index, schema in enumerate(self.schemas):
            sub_path = self._sub_path(current_path, index)
            value = schema.from_url_params(sub_path, params)
            if value is NoDefault:
                raise SchemaError(current_path, "missing_value")
            values.append(value)
        return values

    def to_url_params(
        self, current_path: str, item: ExternalType
    ) -> Iterator[Tuple[str, str]]:
        for index, schema in enumerate(self.schemas):
            sub_path = self._sub_path(current_path, index)
            yield from schema.to_url_params(sub_path, item[index])

    @staticmethod
    def _sub_path(current_path: str, index: int):
        sub_path = f"{current_path}.{index}" if current_path else str(index)
        return sub_path

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        properties = (
            (f"t{index}", schema) for index, schema in enumerate(self.schemas)
        )
        attributes = build_attributes(properties, existing_types, object_wrapper)
        type_ = object_wrapper(type("Tuple", (), attributes))
        return type_



import importlib
import pkgutil
from dataclasses import dataclass, field
from typing import List, Dict, Type, TypeVar, Optional

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.utils import resolve_forward_refs

from schemey.json_schema_context import JsonSchemaContext
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.obj_schema import ObjSchema

T = TypeVar("T")


@dataclass
class SchemaContext:
    schema_loaders: List[SchemaLoaderABC] = field(default_factory=list)
    schema_factories: List[SchemaFactoryABC] = field(default_factory=list)
    schemas: Dict[Type, SchemaABC] = field(default_factory=dict)
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)

    def __post_init__(self):
        self.schema_factories.sort()

    def get_obj_schema(self, item_type: Type[T]) -> ObjSchema:
        schema = self.get_schema(item_type)
        schema = ObjSchema(schema, self.marshaller_context.get_marshaller(item_type))
        return schema

    def get_schema(self, item_type: Type[T]) -> SchemaABC:
        item_type = resolve_forward_refs(item_type)
        schema = self.schemas.get(item_type)
        if schema:
            return schema
        json_schema_context = JsonSchemaContext(
            schemas={**self.schemas}, factories=self.schema_factories
        )
        schema = json_schema_context.get_schema(item_type)
        schema = schema.simplify()
        self.schemas[item_type] = schema
        return schema

    def register_schema(self, item_type: Type, schema: SchemaABC):
        item_type = resolve_forward_refs(item_type)
        self.schemas[item_type] = schema

    def register_factory(self, schema_factory: SchemaFactoryABC):
        self.schema_factories.append(schema_factory)
        self.schema_factories.sort(reverse=True)

    def register_loader(self, jsonifier: SchemaLoaderABC):
        self.schema_loaders.append(jsonifier)
        self.schema_loaders.sort(reverse=True)


_default_context = None
CONFIG_MODULE_PREFIX = "schemey_config_"


def get_default_schema_context() -> SchemaContext:
    global _default_context
    if not _default_context:
        _default_context = new_default_schema_context()
    return _default_context


def new_default_schema_context(
    marshaller_context: Optional[MarshallerContext] = None,
) -> SchemaContext:
    if marshaller_context is None:
        marshaller_context = get_default_context()
    context = SchemaContext(marshaller_context=marshaller_context)
    # Set up context based on naming convention
    module_info = (
        m for m in pkgutil.iter_modules() if m.name.startswith(CONFIG_MODULE_PREFIX)
    )
    modules = [importlib.import_module(m.name) for m in module_info]
    modules.sort(key=lambda m: m.priority, reverse=True)
    for m in modules:
        getattr(m, "configure")(context)
    return context


def schema_for_type(
    type_: Type[T], context: Optional[SchemaContext] = None
) -> ObjSchema:
    if context is None:
        context = get_default_schema_context()
    schema = context.get_obj_schema(type_)
    return schema



from typing import Optional

from marshy.types import ExternalItemType

from schemey.any_of_schema import AnyOfSchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


class AnyOfSchemaLoader(SchemaLoaderABC):
    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        any_of = item.get("anyOf")
        if any_of is None:
            return
        schemas = tuple(json_context.load(a) for a in any_of)
        loaded = AnyOfSchema(
            **dict(
                schemas=schemas,
                name=item.get("name"),
                description=item.get("description"),
            )
        )
        return loaded



from dataclasses import dataclass

_SchemaABC = "schemey.schema_abc.SchemaABC"
_JsonSchemaContext = "schemey.json_schema_context.JsonSchemaContext"


@dataclass
class ParamSchema:
    name: str
    schema: _SchemaABC
    required: bool = True
    param_in: str = "query"

    def dump_json(self, json_context: _JsonSchemaContext):
        return {
            "in": self.param_in,
            "name": self.name,
            "schema": self.schema.dump_json_schema(json_context),
            "required": self.required,
        }



from dataclasses import dataclass
from typing import Optional

from marshy.types import ExternalItemType

from schemey.const_schema import ConstSchema
from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext


@dataclass
class ConstSchemaLoader(SchemaLoaderABC):
    priority: int = 10

    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if "const" not in item:
            return None
        schema = ConstSchema(item["const"], item.get("description"))
        return schema



from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Iterator, Set, Dict, Any, Type, Callable

from marshy import ExternalType
from marshy.types import ExternalItemType

from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError
from schemey.str_param_schema_abc import StrParamSchemaABC


@dataclass(frozen=True)
class EnumSchema(StrParamSchemaABC):
    name: str
    enum: Set[ExternalType]
    description: Optional[str] = None

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if item not in self.enum:
            yield SchemaError(current_path or [], "value_not_permitted", item)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        dumped = dict(name=self.name, enum=list(self.enum))
        if self.description:
            dumped["description"] = self.description
        return dumped

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        type_ = existing_types.get(self.name)
        if type_:
            return type_
        attributes = {str(e): e for e in self.enum}
        type_ = Enum(self.name, attributes)
        existing_types[self.name] = type_
        return type_



import inspect
from dataclasses import dataclass
from typing import Type, Optional

from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller_context import MarshallerContext

from schemey.schema_abc import SchemaABC
from schemey.schema_marshaller import SchemaMarshaller
from schemey.schema_context import SchemaContext, get_default_schema_context


@dataclass(frozen=True)
class SchemaMarshallerFactory(MarshallerFactoryABC):
    priority: int = 200
    schemey_context: Optional[SchemaContext] = None

    def create(
        self, context: MarshallerContext, type_: Type
    ) -> Optional[MarshallerABC]:
        if inspect.isclass(type_) and issubclass(type_, SchemaABC):
            schemey_context = self.schemey_context or get_default_schema_context()
            return SchemaMarshaller(schemey_context=schemey_context)



from dataclasses import dataclass, field, is_dataclass, fields, MISSING
from typing import Optional, List, Iterator, Dict, Set, Tuple, Union, Type, Any

import typing

import typing_inspect
from marshy.factory.optional_marshaller_factory import get_optional_type
from marshy.types import ExternalItemType, ExternalType

from schemey._util import filter_none
from schemey.optional_schema import NoDefault
from schemey.param_schema import ParamSchema
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass(frozen=True)
class ObjectSchema(SchemaABC):
    properties: Dict[str, SchemaABC]
    name: Optional[str] = None
    additional_properties: bool = False
    required: Optional[Set[str]] = None
    description: Optional[str] = None

    def get_schema_errors(
        self, item: ExternalItemType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if not isinstance(item, dict):
            yield SchemaError(current_path, "type", item)
            return
        keys = set(item.keys())
        if self.required:
            missing = self.required - keys
            if missing:
                yield SchemaError(
                    current_path, "missing_properties", ", ".join(missing)
                )
        for key, value in item.items():
            property_schema = self.properties.get(key)
            if property_schema:
                keys.remove(key)
                yield from property_schema.get_schema_errors(value, current_path)
        if keys and not self.additional_properties:
            yield SchemaError(current_path, "additional_properties", ", ".join(keys))

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        dumped = filter_none(
            dict(
                type="object",
                name=self.name,
                additionalProperties=self.additional_properties,
                description=self.description,
            )
        )
        properties = {
            k: s.dump_json_schema(json_context) for k, s in self.properties.items()
        }
        if properties:
            dumped["properties"] = properties
        if self.required:
            required = list(self.required)
            required.sort()
            dumped["required"] = required
        return dumped

    def simplify(self) -> SchemaABC:
        properties = {k: p.simplify() for k, p in self.properties.items()}
        schema = ObjectSchema(**{**self.__dict__, "properties": properties})
        return schema

    def get_param_schemas(self, current_path: str) -> Optional[List[ParamSchema]]:
        param_schemas = []
        for name, schema in self.properties.items():
            sub_path = self._sub_path(current_path, name)
            sub_schemas = schema.get_param_schemas(sub_path)
            if sub_schemas is None:
                return None
            param_schemas.extend(sub_schemas)
        return param_schemas

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> Union[ExternalType, type(MISSING), NoDefault]:
        result = {}
        for key, schema in self.properties.items():
            sub_path = self._sub_path(current_path, key)
            value = schema.from_url_params(sub_path, params)
            if value is NoDefault:
                return NoDefault
            elif value is not MISSING:
                result[key] = value
        return result

    def to_url_params(
        self, current_path: str, item: ExternalItemType
    ) -> Iterator[Tuple[str, str]]:
        for key, schema in self.properties.items():
            if key in item:
                sub_path = self._sub_path(current_path, key)
                yield from schema.to_url_params(sub_path, item[key])

    @staticmethod
    def _sub_path(current_path: str, key: str):
        sub_path = f"{current_path}.{key}" if current_path else key
        return sub_path

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: typing.Callable
    ) -> Type:
        type_ = existing_types.get(self.name)
        if type_:
            return type_
        existing_types[self.name] = typing.ForwardRef(self.name)
        attributes = build_attributes(
            self.properties.items().__iter__(), existing_types, object_wrapper
        )
        type_ = object_wrapper(type(self.name, (), attributes))
        existing_types[self.name] = type_
        _resolve_forward_refs(type_, existing_types)
        return type_


def build_attributes(
    properties: Iterator[Tuple[str, SchemaABC]],
    existing_types: Dict[str, Any],
    object_wrapper: typing.Callable,
):
    annotations = {}
    attrs = {"__annotations__": annotations}
    default_attrs = {}
    for name, schema in properties:
        type_ = schema.get_normalized_type(existing_types, object_wrapper)
        if hasattr(schema, "default"):
            if schema.default is NoDefault:
                default = None
                type_ = Optional[get_optional_type(type_) or type_]
            else:
                default = schema.default
            default_attrs[name] = field(default=default)
        else:
            attrs[name] = field()
        annotations[name] = type_
    attrs.update(**default_attrs)
    return attrs


def _resolve_forward_refs(type_, existing_types: Dict[str, Any]) -> typing.Type:
    if isinstance(type_, typing.ForwardRef):
        type_ = existing_types[type_.__forward_arg__]
        return type_
    if is_dataclass(type_):
        for f in fields(type_):
            f.type = type_.__annotations__[f.name] = _resolve_forward_refs(
                f.type, existing_types
            )
        return type_
    origin = typing_inspect.get_origin(type_)
    if origin is not None:
        type_.__args__ = tuple(
            _resolve_forward_refs(a, existing_types) for a in type_.__args__
        )
    return type_



from abc import abstractmethod, ABC
from functools import total_ordering
from typing import TypeVar, Type, Optional

from schemey.schema_abc import SchemaABC

T = TypeVar("T")
NONE_TYPE = type(None)
_JsonSchemaContext = "schemey.json_schema_context.JsonSchemaContext"


@total_ordering
class SchemaFactoryABC(ABC):
    @property
    def priority(self):
        return 100

    @abstractmethod
    def create(
        self, type_: Type, json_context: _JsonSchemaContext
    ) -> Optional[SchemaABC]:
        """
        Create a schema for the type given, or return None if that was not possible
        """

    def __ne__(self, other):
        return self.priority != getattr(other, "priority", None)

    def __lt__(self, other):
        return self.priority < getattr(other, "priority", None)



from dataclasses import dataclass
from typing import Optional

from marshy.types import ExternalItemType

from schemey.loader.schema_loader_abc import SchemaLoaderABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.tuple_schema import TupleSchema


@dataclass
class TupleSchemaLoader(SchemaLoaderABC):
    priority: int = 10

    def load(
        self, item: ExternalItemType, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        if (
            item.get("type") != "array"
            or item.get("prefixItems") is None
            or item.get("items") is not False
        ):
            return None
        schemas = tuple(json_context.load(s) for s in item.get("prefixItems"))
        loaded = TupleSchema(schemas, description=item.get("description"))
        return loaded



from dataclasses import dataclass
from typing import Optional, List, Iterator, Dict, Any, Type, Set, Callable

from marshy.types import ExternalItemType

from schemey._util import filter_none
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_abc import SchemaABC
from schemey.schema_error import SchemaError


@dataclass(frozen=True)
class ArraySchema(SchemaABC):
    item_schema: Optional[SchemaABC] = None
    min_items: int = 0
    max_items: Optional[int] = None
    uniqueness: bool = False
    description: str = None

    def get_schema_errors(
        self, item: List[ExternalItemType], current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if current_path is None:
            current_path = []
        if not isinstance(item, list):
            yield SchemaError(current_path, "type", item)
            return
        if self.item_schema is not None:
            for index, i in enumerate(item):
                current_path.append(str(index))
                yield from self.item_schema.get_schema_errors(i, current_path)
                current_path.pop()
        if self.min_items is not None and len(item) < self.min_items:
            yield SchemaError(current_path, "min_items", item)
        if self.max_items is not None and len(item) >= self.max_items:
            yield SchemaError(current_path, "max_items", item)
        if self.uniqueness is True:
            existing = set()
            for index, i in enumerate(item):
                if i in existing:
                    current_path.append(str(index))
                    yield SchemaError(current_path, "non_unique", i)
                    current_path.pop()
                    return
                existing.add(i)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        items = (
            self.item_schema.dump_json_schema(json_context)
            if self.item_schema
            else None
        )
        dumped = filter_none(
            dict(
                type="array",
                items=items,
                maxItems=self.max_items,
                description=self.description,
            )
        )
        if self.min_items:
            dumped["minItems"] = self.min_items
        if self.uniqueness:
            dumped["uniqueness"] = True
        return dumped

    def simplify(self) -> SchemaABC:
        item_schema = self.item_schema.simplify()
        schema = ArraySchema(**{**self.__dict__, "item_schema": item_schema})
        return schema

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        item_type = self.item_schema.get_normalized_type(existing_types, object_wrapper)
        if self.uniqueness:
            return Set[item_type]
        else:
            return List[item_type]



from dataclasses import dataclass
from typing import Optional, List

import marshy

from schemey.schema_context import schema_for_type

_Node = f"{__name__}.Node"


@dataclass
class Node:
    id: str
    children: Optional[List[_Node]] = None


schema = schema_for_type(Node)
schema_json = marshy.dump(schema.json_schema)
assert schema_json == {
    "$ref": "#$defs/Node",
    "$defs": {
        "Node": {
            "type": "object",
            "name": "Node",
            "additionalProperties": False,
            "properties": {
                "id": {"type": "string"},
                "children": {
                    "type": "array",
                    "items": {"$ref": "#$defs/Node"},
                    "default": None,
                },
            },
            "required": ["id"],
        }
    },
}



from dataclasses import dataclass
from typing import Optional, List, Iterator, Any, Dict, Type, Callable

from marshy.types import ExternalItemType

from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_abc import SchemaABC
from schemey.schema_error import SchemaError

_instance = None


@dataclass(frozen=True)
class NullSchema(SchemaABC):
    def __new__(cls, *args, **kwargs):
        global _instance
        if not _instance:
            _instance = super(NullSchema, cls).__new__(cls, *args, **kwargs)
        return _instance

    def get_schema_errors(
        self, item, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if item is not None:
            yield SchemaError(current_path or [], "type", item)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        dumped = dict(type="null")
        return dumped

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        return type(None)



from abc import abstractmethod, ABC
from dataclasses import MISSING
from typing import (
    Iterator,
    Optional,
    List,
    TypeVar,
    Dict,
    Tuple,
    Union,
    Any,
    Type,
    Callable,
)

from marshy.types import ExternalItemType, ExternalType

from schemey.schema_error import SchemaError

T = TypeVar("T")
_SchemaABC = f"{__name__}SchemaABC"
_JsonSchemaContext = "schemey.json_schema_context.JsonSchemaContext"
_ParamSchema = "schemey.param_schema.ParamSchema"
# Alias so type checker will stop warning about unimplemented methods (implementation is controlled by whether the
# get_params_schemas returns results or None)
_NotImplementedAlias = NotImplementedError


class NoDefault:
    pass


class SchemaABC(ABC):
    """
    A Schema for a particular type of object, which may be marshalled into a JsonSchema. Json Schemas are fundamentally
    extensible and tolerant of additional unknown attributes. We use this to pass additional data to clients and store
    things that may not be part of the general spec.
    """

    @abstractmethod
    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        """Get the validation errors for the item given."""

    def validate(self, item: ExternalType, current_path: Optional[List[str]] = None):
        """Validate the item given"""
        errors = self.get_schema_errors(item, current_path)
        error = next(errors, None)
        if error:
            raise error

    @abstractmethod
    def dump_json_schema(self, json_context: _JsonSchemaContext) -> ExternalItemType:
        """Convert this to a json schema"""

    def simplify(self) -> _SchemaABC:
        return self

    def get_param_schemas(self, current_path: str) -> Optional[List[_ParamSchema]]:
        """
        Get OpenAPI json schemas for this schema. Not all schemas can be processed as url parameters
        some more complicated schemas are better processed as json input, and will return None
        """
        return None

    def from_url_params(
        self, current_path: str, params: Dict[str, List[str]]
    ) -> Union[ExternalType, type(MISSING), NoDefault]:
        """
        Convert the url params given to a json item. MISSING Implies no parameter was available, but a default is.
        NoDefault implies no parameter was available and there was no default possible
        """
        return NoDefault

    def to_url_params(
        self, current_path: str, item: ExternalType
    ) -> Iterator[Tuple[str, str]]:
        """Convert the item given to url params Raise NotImplemented if get_param_schema is none."""
        raise _NotImplementedAlias()

    @abstractmethod
    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        """
        Build a standard mutable primitive dataclass / list from the externalized version of this schema.
        This is used to specify types to other python systems that do not support json schema, and to
        specify types for graphql
        One major thing to note about of the standardized form is that it does not support dynamic defaults,
        so MISSING and NONE are effectively the same thing, and there is no such thing as Tuples (because graphql
        does not support them)
        """



from typing import Type, Optional

import typing_inspect

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.tuple_schema import TupleSchema


class TupleSchemaFactory(SchemaFactoryABC):
    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        origin = typing_inspect.get_origin(type_)
        if origin is tuple:
            args = typing_inspect.get_args(type_)
            if len(args) == 2 and args[1] is Ellipsis:
                return
            schema = TupleSchema(tuple(json_context.get_schema(s) for s in args))
            return schema



from dataclasses import dataclass
from typing import Optional, List, Iterator, Dict, Any, Type, Callable

from marshy import ExternalType
from marshy.types import ExternalItemType

from schemey.schema_abc import SchemaABC, _JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass
class ConstSchema(SchemaABC):
    """Schema representing a single constant value"""

    const: ExternalType
    description: Optional[str] = None

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        if item != self.const:
            yield SchemaError(current_path, "invalid_value", item)

    def dump_json_schema(self, json_context: _JsonSchemaContext) -> ExternalItemType:
        dumped = dict(const=self.const)
        if self.description:
            dumped["description"] = self.description
        return dumped

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        return type(self.const)



from dataclasses import dataclass, field
from typing import Optional, List, Iterator, Dict, Any, Callable, Type

from marshy import ExternalType
from marshy.types import ExternalItemType

from schemey.schema_abc import SchemaABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_error import SchemaError


@dataclass
class DeferredSchema(SchemaABC):
    ref: str
    schema: Optional[SchemaABC] = field(default=None, compare=False)
    num_usages: int = 0

    def get_schema_errors(
        self, item: ExternalType, current_path: Optional[List[str]] = None
    ) -> Iterator[SchemaError]:
        yield from self.schema.get_schema_errors(item, current_path)

    def dump_json_schema(self, json_context: JsonSchemaContext) -> ExternalItemType:
        json_context.defs[self.ref] = self
        dumped = {"$ref": f"{json_context.defs_path}/{self.ref}"}
        return dumped

    def simplify(self) -> SchemaABC:
        if self.num_usages >= 2:
            return self
        self.num_usages = 0
        return self.schema.simplify()

    def get_normalized_type(
        self, existing_types: Dict[str, Any], object_wrapper: Callable
    ) -> Type:
        return self.schema.get_normalized_type(existing_types, object_wrapper)



from typing import Optional, Type

from schemey.factory.schema_factory_abc import SchemaFactoryABC
from schemey.json_schema_context import JsonSchemaContext
from schemey.schema_abc import SchemaABC

SCHEMA_FACTORY = "__schema_factory__"


class FactorySchemaFactory(SchemaFactoryABC):
    priority: int = 110

    def create(
        self, type_: Type, json_context: JsonSchemaContext
    ) -> Optional[SchemaABC]:
        factory = getattr(type_, SCHEMA_FACTORY, None)
        if factory is not None:
            return factory(json_context)



from dataclasses import dataclass, field
from typing import Optional
from schemey.number_schema import NumberSchema
from schemey.schema_context import schema_for_type


@dataclass
class Person:
    name: str
    age: Optional[int] = field(
        default=None, metadata=dict(schemey=NumberSchema(minimum=0))
    )


schema = schema_for_type(Person)

errors = list(
    schema.get_schema_errors(Person("You", -1))
)  # [SchemaError(path='age', code='minimum', value=-1)]
